#ifndef DUPLEX_H
#define DUPLEX_H

#include "encoder.h"
#include "bdd.h"

#include <deque>
#include <unordered_map>

namespace SATABP {

// Minisat+: Eén, N., Sörensson, N.: Translating pseudo-boolean constraints into SAT, JSAT 2006
// BDD1: Abio, I., Nieuwenhuis, R., Oliveras, A., Rodriguez-Carbonell, E., Mayer-Eichberger, V.: A new look at BDDs for pseudo-boolean constraints, JAIR 2012
enum NodeEncoderType {
    minisat,        //Minisat+ encoding without red-clauses
    minisat_gac,    //Minisat+ encoding with red-clauses
    bdd1,            //BDD1
    cpaior,        //Encoding used in the experiments of CPAIOR'20 submission, NOT THE BEST
};

class DuplexEncoder : public Encoder {
public:
  DuplexEncoder(Graph* g, ClauseContainer* cc, VarHandler* vh, NodeEncoderType node_cnf);
  virtual ~DuplexEncoder();
private:
  BDDHandler bh;
  NodeEncoderType node_encoder = minisat_gac;

  std::unordered_map<int,std::vector<int>> fwd_amo_roots;
  std::unordered_map<int,std::vector<int>> bwd_amo_roots;
  std::unordered_map<int,std::vector<int>> fwd_amz_roots;
  std::unordered_map<int,std::vector<int>> bwd_amz_roots;

  std::unordered_map<int,std::vector<std::vector<int>>> node_amz_literals;
  unsigned number_of_windows;
  unsigned last_window_w;


  void init_members();

  void do_encode_antibandwidth(unsigned w, const std::vector<std::pair<int,int>>& node_pairs) final;

  int do_vars_size() const final;

  void construct_window_bdds(int w);
  void construct_node_window_bdds(int w, unsigned var_group);
  void glue_window_bdds();
  void glue_node_window_bdds(unsigned var_group);

  void glue_edge_windows(int node1, int node2);
  void make_equal_bdds(BDD_id b1, BDD_id b2);
  void encode_column_eo();
  void seq_encode_column_eo();

  void product_encode_eo(const std::vector<int>& vars);
  void product_encode_amo(const std::vector<int>& vars);
  void seq_encode_amo(const std::vector<int>& vars);

  // Not & on purpose!
  BDD_id build_amo(std::deque<unsigned int> vars);
  BDD_id build_amz(std::deque<unsigned int> vars);

  void add_amo_clauses(int node_id, int dec_var, BDD_id true_child, BDD_id false_child, int vars_size);
  void add_amz_clauses(int node_id, int dec_var,BDD_id false_child);

  unsigned measure_scamo_encoding_size(unsigned w) final;
};

}

#endif
